r"""
Moments of a Compound Poisson Process variable

Compound Poisson Process

.. math::

   J(t) = \sum_{i=0}^{N(t)}j_i,\quad j_i \sim \mathcal{N}(\mu_j,\sigma_j^2)

where :math:`N(t)` is a Poisson process with rate :math:`\lambda`.
For our purpose, I define variable
:math:`J_n \triangleq \sum_{i=N((n-1)h)+1}^{N(nh)}j_i` if
:math:`N(nh) - N((n-1)h) > 0`, otherwise :math:`J_n \triangleq 0`.

Moment-Generating Function
===========================

For variable :math:`J_n`, its moment-generating function

.. math::

   M_{J_n}(s) = e^{\lambda h (M_{j_i}(s)-1)},
   \qquad
   M_{j_i}(s) = e^{\mu_js+\frac{1}{2}\sigma_j^2s^2}.

where :math:`M_{j_i}(s)` is the moment-generating function of normal variable
:math:`j_i`.

MGF - :abbr:`CPP(Compound Poisson Process)`
--------------------------------------------

For the first three derivatives,

.. math::

   M_{J_n}^{'}(s)
   &= e^{\lambda h (M_{j_i}(s)-1)} (\lambda h) M_{j_i}^{'}(s),\\
   M_{J_n}^{''}(s)
   &= e^{\lambda h (M_{j_i}(s)-1)} \left[ (\lambda h)^2 M_{j_i}^{'2}(s)
   + (\lambda h) M_{j_i}^{''}(s) \right],\\
   M_{J_n}^{(3)}(s)
   &= e^{\lambda h (M_{j_i}(s)-1)} \left[ (\lambda h)^3 M_{j_i}^{'3}(s)
   + 3(\lambda h)^2 M_{j_i}^{'}(s) M_{j_i}^{''}(s)
   + (\lambda h) M_{j_i}^{(3)}(s) \right].

I propose to represent derivative of any order as

.. math::

   M^{(n)}_{J_n}(s)
   = \sum_{(i,(n1,m1),...,(nl,ml))} b_{key} (\lambda h)^i M_{j_i}^{(n1)m1}(s)
   \cdots M_{j_i}^{(nl)ml}(s)

where the leading term :math:`e^{\lambda h (M_{j_i}(s)-1)}` has been omitted
for notation simplicity,
:math:`n1 < \cdots < nl` and :math:`key=(i,(n1,m1),...,(nl,ml))`.
Then its derivative can be computed as

.. math::

   M_{J_n}^{(n+1)}(s)
   &= \sum_{(i,(n1,m1),...,(nl,ml))} b_{key} (\lambda h)^{i+1} M_{j_i}^{(1)}(s)
   M_{j_i}^{(n1)m1}(s)\cdots M_{j_i}^{(nl)ml}(s)\\
   &\quad+ \sum_{(i,(n1,m1),...,(nl,ml))} b_{key} (\lambda h)^i
   (M_{j_i}^{(n1)m1}(s)\cdots M_{j_i}^{(nl)ml}(s))',

where again the leading term :math:`e^{\lambda h (M_{j_i}(s)-1)}` has also
been omitted.
Rearrage the derivative to represent it as that of :math:`M^{(n)}_{J_n}(s)`.


MGF - Normal Distribution
--------------------------

For the first three derivatives,

.. math::

   M_{j_i}^{'}(s)
   &= e^{\mu_js+\frac{1}{2}\sigma_j^2s^2} (\mu_j + \sigma_j^2s),\\
   M_{j_i}^{''}(s)
   &= e^{\mu_js+\frac{1}{2}\sigma_j^2s^2} \left[ (\mu_j + \sigma_j^2s)^2
   +  \sigma_j^2 \right],\\
   M_{j_i}^{(3)}(s)
   &= e^{\mu_js+\frac{1}{2}\sigma_j^2s^2} \left[ (\mu_j + \sigma_j^2s)^3
   +  (\mu_j + \sigma_j^2s)\sigma_j^2 +
   2(\mu_j + \sigma_j^2s)\sigma_j^2\right].

I propose to represent derivative of any order as

.. math::

   M_{j_i}^{(n)}(s)
   = \sum_{i,j}b_{ij}(\mu+\sigma^2s)^i \sigma^{2j},

where the leading term :math:`e^{\mu_js+\frac{1}{2}\sigma_j^2s^2}` has
been omitted for notation simplicity.

Then its derivative is given as

.. math::

   M_{j_i}^{(n+1)}(s)
   = \sum_{i,j}b_{ij}(\mu+\sigma^2s)^{i+1} \sigma^{2j}
    +\sum_{i>0,j}b_{ij}i(\mu+\sigma^2s)^{i-1} \sigma^{2(j+1)}.

Rearrange the derivative to represent it as that of :math:`M^{(n)}_{j_i}(s)`.

In summary, I defined

1. :py:func:`~ajdmom.cpp_mom.mnorm` to compute moment of normal variable
   which uses
   :py:func:`~ajdmom.cpp_mom.dmgf`.

2. :py:func:`~ajdmom.cpp_mom.mcpp` to compute moment of CPP variable
   which uses
   :py:func:`~ajdmom.cpp_mom.dmgf_cpp`,
   :py:func:`~ajdmom.cpp_mom.decode`.

3. :py:func:`~ajdmom.cpp_mom.cmcpp` to compute central moment of CPP
   variable.
"""
from fractions import Fraction as Frac
import math

from ajdmom.poly import Poly


#
# Normal distribution
#

def dmgf(poly):
    """Derivative of normal Moment-Generating Function

    :param Poly poly: poly with attribute ``keyfor`` = ('mu','sigma^2').
    :return: poly with attribute ``keyfor`` = ('mu','sigma^2').
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(['mu', 'sigma^2'])
    for k in poly:
        # step 1
        key = (k[0] + 1, k[1])
        poln.add_keyval(key, poly[k])
        # step 2
        if k[0] > 0:
            knw = (k[0] - 1, k[1] + 1)
            poln.add_keyval(knw, poly[k] * k[0])
    return poln


def mnorm(n):
    """Moment of Normal distribution

    :param int n: order of the moment, n >= 0.
    :return: poly with attribute ``keyfor`` = ('mu','sigma^2').
    :rtype: Poly
    """
    if n < 0:
        msg = f'mnorm(n): order of the moment is {n} which must be > 0!'
        raise ValueError(msg)
    M = []
    kf = ['mu', 'sigma^2']
    # n = 0
    poly = Poly({(0, 0): 1})
    poly.set_keyfor(kf)
    M.append(poly)
    #
    # special case
    #
    if n == 0: return poly
    #
    # typical cases
    #
    # n = 1
    poly = Poly({(1, 0): 1})
    poly.set_keyfor(kf)
    M.append(poly)
    if n == 1: return poly
    #
    for i in range(2, n + 1):
        M.append(dmgf(M[-1]))
    return M[-1]


#
# Compound Poisson Process
#

def d1_times_key(key):
    r"""Update the key after multiply with :math:`(\lambda h)M_{j_i}'(s)`

    :param tuple key: a tuple of (i,(n1,m1),...,(nl,ml)) which
       stands for :math:`(\lambda h)^{i} M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)`.
    :return: updated key.
    :rtype: tuple
    """
    # multiply with (lambda h)
    knw = list(key)
    knw[0] += 1
    # multiply with M_{j_i}'(s)
    if key[1][0] == 1:  # (n1,m1),(n2,m2),...,(nj,mj) where n1 < n2 < ... < nj
        knw[1] = (1, key[1][1] + 1)
    else:
        knw.insert(1, (1, 1))
    return tuple(knw)


def dterm(key, coef):
    r"""Take derivative of each term

    :param tuple key: key corresponding to a term, i.e.,
       (i,(n1,m1),(n2,m2),...,(nj,mj)) which
       stands for :math:`(\lambda h)^{i} M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)`.
    :param Fraction coef: corresponding coefficient.
    :return: poly representing the derivative.
    :rtype: Poly
    """
    poly = Poly()
    # key: tuple(i,(n1,m1),(n2,m2),...,(nj,mj)) where n1 < n2 < ... < nj
    for j in range(1, len(key)):
        knw = list(key)
        n_j, m_j = key[j]
        coef_new = coef * m_j  # update coef
        knw[j] = (n_j, m_j - 1)  # update exponent M^{(n_j)m_j} -> M^{(n_j)(m_j-1)}
        #
        if j < len(key) - 1:  # having next one
            n_nxt, m_nxt = key[j + 1]
            if n_nxt == n_j + 1:
                knw[j + 1] = (n_nxt, m_nxt + 1)
            else:
                knw.insert(j + 1, (n_j + 1, 1))
        else:  # no follower
            knw.append((n_j + 1, 1))
        #
        if knw[j][1] == 0: del knw[j]  # delete M^{(n_j)0}
        poly.add_keyval(tuple(knw), coef_new)
    return poly


def dmgf_cpp(poly):
    r"""derivative of Moment Generating Function of CPP

    :param Poly poly: poly representing MGF with
       key :math:`(i,(n_1,m_1),...,(n_l,m_l))` which
       stands for :math:`(\lambda h)^{i}M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)`.
    :return: poly with the same key-val.
    :rtype: Poly
    """
    # step 1
    poly_sum = Poly()
    for key in poly:
        knw = d1_times_key(key)
        poly_sum.add_keyval(knw, poly[key])
    # step 2
    for key in poly:
        poly_sum.merge(dterm(key, poly[key]))
    return poly_sum


def decode(poly):
    r"""Decode intermediate poly to target poly

    :param Poly poly: poly representation of MGF with
       key :math:`(i,(n_1,m_1),...,(n_l,m_l))` which
       stands for :math:`(\lambda h)^{i} M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)`.
    :return: poly with attribute ``keyfor`` = ('lambda*h','mu','sigma^2').
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(['lambda*h', 'mu', 'sigma^2'])
    for k in poly:
        i = k[0]
        #
        pol1 = Poly({(0, 0): 1})
        pol1.set_keyfor(['mu', 'sigma^2'])
        for j in range(1, len(k)): # may error: len(k) -> len(k) + 1: no error
            n_j, m_j = k[j]
            pol1 = pol1 * (mnorm(n_j) ** m_j)
        #
        for key in pol1:
            knw = (i, key[0], key[1])
            val = poly[k] * pol1[key]
            poln.add_keyval(knw, val)
    poln.remove_zero()
    return poln


def mcpp_original(n):
    """Moment of Compound Poisson Process variable of order n

    :param int n: order of the moment.
    :return: poly with attribute ``keyfor`` = ('lambda*h','mu','sigma^2').
    :rtype: Poly
    """
    if n < 0:
        raise ValueError(f'mcpp(n): order of the moment is {n} which must > 0!')
    M = []
    kf = ['(lambda h)^{i} M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)']
    # n = 0
    poly = Poly({(0,): 1})
    poly.set_keyfor(kf)
    M.append(poly)
    #
    if n == 0:
        poln = Poly({(0, 0, 0): 1})
        poln.set_keyfor(['lambda*h', 'mu', 'sigma^2'])
        return poln
    # n = 1
    poly = Poly({(1, (1, 1)): 1})
    poly.set_keyfor(kf)
    M.append(poly)
    #
    if n == 1:
        poln = Poly({(1, 1, 0): 1})
        poln.set_keyfor(['lambda*h', 'mu', 'sigma^2'])
        return poln
    #
    for i in range(2, n + 1):
        M.append(dmgf_cpp(M[-1]))
    return decode(M[-1])


def mcpp(n):
    """Moment of Compound Poisson Process variable of order n

    :param int n: order of the moment.
    :return: poly with attribute ``keyfor`` = ('lambda*h','mu','sigma').
    :rtype: Poly
    """
    poly = mcpp_original(n)  # ('lambda*h','mu','sigma^2')
    poln = Poly()
    poln.set_keyfor(['lambda*h', 'mu', 'sigma'])
    for k in poly:
        key = (k[0], k[1], 2 * k[2])
        val = poly[k]
        poln.add_keyval(key, val)
    return poln


def cmcpp(n):
    """Central Moment of Compound Poisson Process of order n

    :param int n: order of the moment.
    :return: poly with attribute ``keyfor`` = ('lambda*h','mu','sigma')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError(f'cmcpp(n): Order of the moment is {n} which must > 0!')
    poly = Poly()
    poly.set_keyfor(['lambda*h', 'mu', 'sigma'])
    for i in range(n + 1):
        coef = math.comb(n, i) * ((-1) ** (n - i))
        poln = mcpp(i)  # ('lambda*h','mu','sigma')
        # * (lambda * h * mu)^{n-i}
        for k in poln:
            key = (k[0] + n - i, k[1] + n - i, k[2])
            val = poln[k] * coef
            poly.add_keyval(key, val)
    poly.remove_zero()
    return poly


if __name__ == '__main__':
    # Example usage of the module, see 'tests/test_cpp_mom.py' for more test
    from pprint import pprint

    print('\nExample usage of the module functions\n')
    # test moment calculation of normal distribution
    kf = ('mu', 'sigma^2')
    msg = f"mnorm(n) returns the nth moment of normal,\na poly with keyfor = {kf}"
    print(msg)
    print('mnorm(8) = ')
    pprint(mnorm(n=8))
    # test moment calculation of Compound Poisson Process variable
    kf = ('lambda*h', 'mu', 'sigma')
    msg = "\nmcpp(n) returns the nth moment of CPP with normal jump, "
    msg += f"\na poly with keyfor = {kf}"
    print(msg)
    print('mcpp(4) = ')
    pprint(mcpp(n=4))
    # test central moment calculation
    kf = ('lambda*h', 'mu', 'sigma')
    msg = "\ncmcpp(n) returns the nth cmoment of CPP with normal jump, "
    msg += f"\na poly with keyfor = {kf}"
    print(msg)
    print('cmcpp(4) = ')
    pprint(cmcpp(n=4))



"""
Moments for Two-Factor SV
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.ito_mom import moment_v
from ajdmom.itos_mom import moment_IEI_IEII, t_mul_t0
from ajdmom.utils import comb
from ajdmom.mdl_2fsv.cmom import (
    vvee_IEI_IEII,
    dt0
)


def b_n(n1, n2, n3, n4, n5, n6, n7):
    """Coefficient corresponding to the combination

    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{1,n-1}` being selected.
    :param int n3: times of :math:`v_{2,n-1}` being selected.
    :param int n4: times of :math:`eI_{1,n}` being selected.
    :param int n5: times of :math:`I_{1,n}` being selected.
    :param int n6: times of :math:`eI_{2,n}` being selected.
    :param int n7: times of :math:`I_{2,n}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('e^{-k1h}','k1^{-}','theta1','sigma_v1',
       'e^{-k2h}','k2^{-}','theta2','sigma_v2', 'h','mu').
    :rtype: Poly
    """
    b = Poly()
    kf = ['e^{-k1h}', 'k1^{-}', 'theta1', 'sigma_v1',
          'e^{-k2h}', 'k2^{-}', 'theta2', 'sigma_v2', 'h', 'mu']
    b.set_keyfor(kf)
    #
    for i1 in range(n1, -1, -1):  # 1/2k1 * theta1
        for i2 in range(n1 - i1, -1, -1):  # -e^{-k1h}/2k1 *theta1
            for i3 in range(n1 - i1 - i2, -1, -1):  # 1/2k2 * theta2
                for i4 in range(n1 - i1 - i2 - i3, -1, -1):  # -e^{-k2h}/2k2 *theta2
                    for i5 in range(n1 - i1 - i2 - i3 - i4, -1, -1):  # -theta1 h/2
                        for i6 in range(n1 - i1 - i2 - i3 - i4 - i5, -1, -1):  # -theta2 h/2
                            i7 = n1 - i1 - i2 - i3 - i4 - i5 - i6  # mu h
                            for j in range(n2 + 1):  # -e^{-k1h}/2k1
                                for q in range(n3 + 1):  # -e^{-k2h}/2k2
                                    num = comb(n1, [i1, i2, i3, i4, i5, i6, i7])
                                    num *= math.comb(n2, j)
                                    num *= math.comb(n3, q)
                                    #
                                    key = (i2 + j, i1 + i2 + n2 + n4 + n5, i1 + i2 + i5, n4 + n5,
                                           i4 + q, i3 + i4 + n3 + n6 + n7, i3 + i4 + i6, n6 + n7,
                                           i5 + i6 + i7, i7)
                                    num *= (-1) ** (i2 + i4 + i5 + i6 + (n2 + j) + (n3 + q) + n5 + n7)
                                    den = 2 ** (n1 - i7 + n2 + n3 + n4 + n5 + n6 + n7)
                                    b.add_keyval(key, Frac(num, den))
    return b


def moment_comb(n, n1, n2, n3, n4, n5, n6, n7, n8):
    """Moment for this combination of given n1 to n8 with n.

    :param int n: l in :math:`E[y_{n}^l]`.
    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{1,n-1}` being selected.
    :param int n3: times of :math:`v_{2,n-1}` being selected.
    :param int n4: times of :math:`eI_{1,n}` being selected.
    :param int n5: times of :math:`I_{1,n}` being selected.
    :param int n6: times of :math:`eI_{2,n}` being selected.
    :param int n7: times of :math:`I_{2,n}` being selected.
    :param int n8: times of :math:`I_{n}^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = vvee_IEI_IEII(n2, n3, n4, n5, n6, n7, n8)
    # ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    #  'e^{-(n1*k1+n2*k2)h}','h',
    #  'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2']
    b = b_n(n1, n2, n3, n4, n5, n6, n7)
    # ['e^{-k1h}','k1^{-}','theta1','sigma_v1',
    #  'e^{-k2h}','k2^{-}','theta2','sigma_v2','h','mu']
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    #
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k1 in poly:
        for k2 in b:
            # t0 = list(k1[0]); t0.insert(0,(1,0,k2[1])); t0.insert(0,(0,1,k2[5]))
            # t0 = tuple(t0)
            t0 = t_mul_t0((1, 0, k2[1]), k1[0])
            t0 = t_mul_t0((0, 1, k2[5]), t0)
            #
            t1 = (k1[1][0] + k2[0], k1[1][1] + k2[4])
            #                   h        mu  v_{1,n-1}  theta1       sigma_v1
            key = (t0, t1, k1[2] + k2[8], k2[9], k1[3], k1[4] + k2[2], k1[5] + k2[3],
                   # v_{2,n-1} theta2     sigma_v2
                   k1[6], k1[7] + k2[6], k1[8] + k2[7])
            val = poly[k1] * b[k2]
            poln.add_keyval(key, val)
    c = comb(n, [n1, n2, n3, n4, n5, n6, n7, n8])
    return c * poln


def sub_v(poly):
    """Substitute :math:`v_{1,n-1}` and :math:`v_{2,n-1}` with their moments

    :param Poly poly: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    poly_sum = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2']
    poly_sum.set_keyfor(kf)
    #
    poln = Poly()
    for k in poly:
        pol1 = moment_v(k[4])  # [theta1, sigma_v1^2/k1]
        for key in pol1:
            i, j = key
            # t0 = list(k[0]); t0.insert(0, (1,0,j)); t0 = tuple(t0)
            t0 = t_mul_t0((1, 0, j), k[0])
            #
            knw = (t0, k[1], k[2], k[3], k[5] + i, k[6] + 2 * j, k[7], k[8], k[9])
            val = poly[k] * pol1[key]
            poln.add_keyval(knw, val)
    for k in poln:
        pol2 = moment_v(k[6])  # [theta2, sigma_v2^2/k2]
        for key in pol2:
            i, j = key
            # t0 = list(k[0]); t0.insert(0, (0,1,j)); t0 = tuple(t0)
            t0 = t_mul_t0((0, 1, j), k[0])
            #
            knw = (t0, k[1], k[2], k[3], k[4], k[5], k[7] + i, k[8] + 2 * j)
            val = poln[k] * pol2[key]
            poly_sum.add_keyval(knw, val)
    return poly_sum


def moment_y(l):
    """Moment of :math:`y_n` with order :math:`l`

    :param int l: order of the moment.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly.set_keyfor(kf)
    #
    for n1 in range(l, -1, -1):
        for n2 in range(l - n1, -1, -1):
            for n3 in range(l - n1 - n2, -1, -1):
                for n4 in range(l - n1 - n2 - n3, -1, -1):
                    for n5 in range(l - n1 - n2 - n3 - n4, -1, -1):
                        for n6 in range(l - n1 - n2 - n3 - n4 - n5, -1, -1):
                            for n7 in range(l - n1 - n2 - n3 - n4 - n5 - n6, -1, -1):
                                n8 = l - n1 - n2 - n3 - n4 - n5 - n6 - n7
                                poly.merge(moment_comb(l, n1, n2, n3, n4, n5, n6, n7, n8))
    poly_sum = sub_v(poly)
    poly_sum.remove_zero()
    return poly_sum


##########
# scalar and (partial) derivative
##########

def dpoly(poly, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param Poly poly: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. k1 or k2
    if wrt == 'k1' or wrt == 'k2':
        for k in poly:
            t0 = k[0]
            if len(t0) != 0:
                deriv = dt0(t0, wrt)
                if len(deriv) != 0:
                    for t in deriv:
                        t0_new, val_new = t
                        knw = list(k)
                        knw[0] = t0_new
                        val = val_new * poly[k]
                        pold.add_keyval(tuple(knw), val)
        for k in poly:
            if wrt == 'k1':
                n = k[1][0]
            else:
                n = k[1][1]
            if n != 0:
                knw = list(k)
                knw[2] += 1
                val = (-n) * poly[k]
                pold.add_keyval(tuple(knw), val)
    # partial derivative w.r.t. mu
    elif wrt in ['mu', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2']:
        if wrt == 'mu': i = 3
        if wrt == 'theta1': i = 4
        if wrt == 'sigma_v1': i = 5
        if wrt == 'theta2': i = 6
        if wrt == 'sigma_v2': i = 7
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                val = k[i] * poly[k]
                pold.add_keyval(tuple(knw), val)
    else:
        candidates = "'k1','k2','mu','theta1','sigma_v1','theta2','sigma_v2'"
        raise ValueError(f"wrt must be one of {candidates}!")
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k1 = par['k1']
    k2 = par['k2']
    h = par['h']
    mu = par['mu']
    theta1 = par['theta1']
    theta2 = par['theta2']
    sigma_v1 = par['sigma_v1']
    sigma_v2 = par['sigma_v2']
    #
    value = 0
    for k in poly:
        val = poly[k]
        t0 = k[0]
        for t in t0:
            val *= (t[0] * k1 + t[1] * k2) ** (-t[2])
        t1 = k[1]
        val *= math.exp(-(t1[0] * k1 + t1[1] * k2) * h)
        val *= h ** k[2]
        val *= mu ** k[3]
        val *= theta1 ** k[4]
        val *= sigma_v1 ** k[5]
        val *= theta2 ** k[6]
        val *= sigma_v2 ** k[7]
        value += val
    return value


def m(l, par):
    """Moment in scalar

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :return: scalar of the moment.
    :rtype: float
    """
    moment = moment_y(l)
    value = poly2num(moment, par)
    return value


def dm(l, par, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    moment = moment_y(l)
    pold = dpoly(moment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_2fsv.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    mom = moment_y(n)
    print(f"moment_y({n}) = \n")
    pprint(mom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{mom.keyfor}")



"""
Conditional Central Moments of the SVCJ model, given initial variance
"""
import math
from fractions import Fraction

from ajdmom import Poly
from ajdmom.mdl_svcj.cond_mom import moment_y, poly2num

def cmoment_y(n):
    r"""conditional central moment of :math:`\overline{y}_t|v_0`

    :param integer n: order of the conditional central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 't', 'k^{-}',
      'mu', 'v0-theta', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s')
    :rtype: Poly
    """
    mean_y = moment_y(1)
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    poly = Poly()
    poly.set_keyfor(kf)
    if n < 0:
        raise ValueError(f'cmoment_y({n}) is called')
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    # special case
    if n == 0:
        return p1
    # typical cases
    for i in range(n + 1):
        bino = math.comb(n, i)
        sign = (-1) ** i
        pol1 = p1 if i == 0 else mean_y ** i
        poln = (bino * sign) * (pol1 * moment_y(n - i))
        poly.merge(poln)
    return poly

def cm_y(n, par):
    cmoment = cmoment_y(n)
    value = poly2num(cmoment, par)
    return value

if __name__ == '__main__':
    from pprint import pprint

    # n = 1
    # poly = cmoment_y(n)
    # print(f"cmoment_y({n}) = "); pprint(poly); print(f"which is a poly with keyfor = {poly.keyfor}")

    v0 = 0.007569
    k = 3.46
    theta = 0.008
    sigma = 0.14
    rho = -0.82
    r = 0.0319
    lmbd = 0.47
    mu_b = -0.1
    sigma_s = 0.0001
    mu_v = 0.05
    rhoJ = -0.38

    mu = r - lmbd * mu_b
    mu_s = math.log((1 + mu_b) * (1 - rhoJ * mu_v)) - sigma_s ** 2 / 2
    h = 1

    par = {'v0': v0, 'mu': mu, 'k': k, 'theta': theta, 'sigma': sigma, 'rho': rho,
           'lmbd': lmbd, 'mu_v': mu_v, 'rhoJ': rhoJ, 'mu_s': mu_s, 'sigma_s': sigma_s,
           'h': h}

    for n in range(1, 4 + 1):
        print(f"cmoment({n}) = {cm_y(n, par)}")



## Moments and Central Moments
from ajdmom.mdl_1fsv.mom import m, dm  # for moments
from ajdmom.mdl_1fsv.cmom import cm, dcm  # for central moments

parameters = {'mu': 0.125, 'k': 0.1, 'theta': 0.25,
              'sigma_v': 0.1, 'rho': -0.7, 'h': 1}

# 3rd moment as an example
moment = m(l=3, par=parameters)  # cm: central moment
# print(f"3rd moment = {moment}")
# partial derivative w.r.t. parameter 'k'
dmoment = dm(l=3, par=parameters, wrt='k')  # dcm: central moment

## Covariance
from ajdmom.mdl_1fsv.cov import cov, dcov

parameters = {'mu': 0.125, 'k': 0.1, 'theta': 0.25,
              'sigma_v': 0.1, 'rho': -0.7, 'h': 1}

# covariance cov(y_n^2, y_{n+1}^2) as an example
covariance = cov(l1=2, l2=2, par=parameters)
# partial derivative w.r.t. parameter 'k'
dcovariance = dcov(l1=2, l2=2, par=parameters, wrt='k')



"""
Module for Moments for Two-Factor SV with jumps
"""
import math

from ajdmom.poly import Poly
from ajdmom.cpp_mom import mcpp
from ajdmom.mdl_2fsv.mom import moment_y as m_y
from ajdmom.mdl_2fsv.cmom import dt0


def moment_y(l):
    """Moment of :math:`y_n` with order :math:`l`

    :param int l: order of the moment.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2', 'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'theta1', 'sigma_v1', 'theta2', 'sigma_v2', 'lambda', 'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    #
    for i in range(l, -1, -1):
        pol1 = m_y(i)
        # keyfor = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
        #   'e^{-(n1*k1+n2*k2)h}','h','mu',
        #   'theta1','sigma_v1', 'theta2','sigma_v2')
        pol2 = mcpp(l - i)
        # keyfor = ('lambda*h','mu_j','sigma_j')
        c = math.comb(l, i)
        for k1 in pol1:
            for k2 in pol2:
                key = (k1[0], k1[1], k1[2] + k2[0], k1[3], k1[4], k1[5], k1[6], k1[7],
                       k2[0], k2[1], k2[2])
                val = c * pol1[k1] * pol2[k2]
                poly.add_keyval(key, val)
    poly.remove_zero()
    return poly


##########
# scalar and (partial) derivative
##########

def dpoly(poly, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param Poly poly: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2','lambda','mu_j','sigma_j').
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2','lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    pold = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'theta1', 'sigma_v1', 'theta2', 'sigma_v2', 'lambda', 'mu_j', 'sigma_j']
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. k1 or k2
    if wrt == 'k1' or wrt == 'k2':
        for k in poly:
            t0 = k[0]
            if len(t0) != 0:
                deriv = dt0(t0, wrt)
                if len(deriv) != 0:
                    for t in deriv:
                        t0_new, val_new = t
                        knw = list(k)
                        knw[0] = t0_new
                        val = val_new * poly[k]
                        pold.add_keyval(tuple(knw), val)
        for k in poly:
            if wrt == 'k1':
                n = k[1][0]
            else:
                n = k[1][1]
            if n != 0:
                knw = list(k)
                knw[2] += 1
                val = (-n) * poly[k]
                pold.add_keyval(tuple(knw), val)
    # partial derivative w.r.t. mu
    elif wrt in ['mu', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2', 'lambda', 'mu_j',
                 'sigma_j']:
        if wrt == 'mu': i = 3
        if wrt == 'theta1': i = 4
        if wrt == 'sigma_v1': i = 5
        if wrt == 'theta2': i = 6
        if wrt == 'sigma_v2': i = 7
        if wrt == 'lambda': i = 8
        if wrt == 'mu_j': i = 9
        if wrt == 'sigma_j': i = 10
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                val = k[i] * poly[k]
                pold.add_keyval(tuple(knw), val)
    else:
        candidates = "'k1','k2','mu','theta1','sigma_v1','theta2','sigma_v2'"
        candidates += ",'lambda','mu_j','sigma_j'"
        raise ValueError(f"wrt must be one of {candidates}!")
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2','lambda','mu_j','sigma_j').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k1 = par['k1']
    k2 = par['k2']
    h = par['h']
    mu = par['mu']
    theta1 = par['theta1']
    theta2 = par['theta2']
    sigma_v1 = par['sigma_v1']
    sigma_v2 = par['sigma_v2']
    lmbd = par['lambda']
    mu_j = par['mu_j']
    sigma_j = par['sigma_j']
    #
    value = 0
    for k in poly:
        val = poly[k]
        t0 = k[0]
        for t in t0:
            val *= (t[0] * k1 + t[1] * k2) ** (-t[2])
        t1 = k[1]
        val *= math.exp(-(t1[0] * k1 + t1[1] * k2) * h)
        val *= h ** k[2]
        val *= mu ** k[3]
        val *= theta1 ** k[4]
        val *= sigma_v1 ** k[5]
        val *= theta2 ** k[6]
        val *= sigma_v2 ** k[7]
        val *= lmbd ** k[8]
        val *= mu_j ** k[9]
        val *= sigma_j ** k[10]
        value += val
    return value


def m(l, par):
    """Moment in scalar

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :return: scalar of the moment.
    :rtype: float
    """
    moment = moment_y(l)
    value = poly2num(moment, par)
    return value


def dm(l, par, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    moment = moment_y(l)
    pold = dpoly(moment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_2fsvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    mom = moment_y(n)
    print(f"moment_y({n}) = \n")
    pprint(mom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{mom.keyfor}")



r"""
Joint Conditional Moment of :math:`I\!E_t,I_t,I_t^{*},I\!E\!Z_t,I\!Z_t,I\!Z_t^{*}|v_0`

In this module, we focus on computing the following conditional joint moment:

.. math::

   \begin{equation}\label{eqn:joint-ieii-ieziziz}
    \mathbb{E}[I\!E_t^{m_1}I_t^{m_2}I_t^{*m_3}I\!E\!Z_t^{m_4}I\!Z_t^{m_5}I\!Z_t^{*m_6}|v_0].
   \end{equation}

While the recursive computation of :math:`\mathbb{E}[I\!E_t^{m_1} I_t^{m_2} I_t^{*m_3}|v_0]` is 
well-established for models such as the Heston model, we encounter a new challenge: 
the last three quantities :math:`I\!E\!Z_t^{m_4} I\!Z_t^{m_5} I\!Z_t^{*m_6}` are not independent
of the first three quantities :math:`I\!E_t^{m_1} I_t^{m_2} I_t^{*m_3}` in above equation.
For :math:`m_4 + m_5 + m_6 \ge 1`, we must evaluate integrals of the form

.. math::

    \begin{equation}
        \int_0^t e^{lks} \mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3} I\!E\!Z_t^{m_4} I\!Z_t^{m_5} I\!Z_t^{*m_6}|v_0]\mathrm{d} s.
    \end{equation}

The dependence between :math:`I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3}` and
:math:`I\!E\!Z_t^{m_4} I\!Z_t^{m_5} I\!Z_t^{*m_6}` motivates us to decompose the latter
as follows:

.. math::

    \begin{align*}
        &I\!E\!Z_t^{m_4}I\!Z_t^{m_5}I\!Z_t^{*m_6}\\
        &= \sum_{i_1=0}^{m_4}\sum_{i_2=0}^{m_5}\sum_{i_3=0}^{m_6} \binom{m_4}{i_1}\binom{m_5}{i_2}
        \binom{m_6}{i_3} I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3} I\!E\!Z_{s,t}^{m_4-i_1}
        I\!Z_{s,t}^{m_5-i_2} I\!Z_{s,t}^{*m_6-i_3}, \quad \forall s \le t,
    \end{align*}

where :math:`I\!E\!Z_t` is split into two independent parts :math:`I\!E\!Z_s, I\!E\!Z_{s,t}`, i.e.,
:math:`I\!E\!Z_t = I\!E\!Z_s + I\!E\!Z_{s,t}`. Similarly, :math:`I\!Z_t` and :math:`I\!Z_t^*`
are decomposed as :math:`I\!Z_t = I\!Z_s + I\!Z_{s,t}` and :math:`I\!Z_t^* = I\!Z_s^* + I\!Z_{s,t}^*`,
respectively. Here, the new terms :math:`I\!E\!Z_{s,t}`, :math:`I\!Z_{s,t}` and :math:`I\!Z_{s,t}^{*}`
are defined as

.. math::

    \begin{equation*}
        I\!E\!Z_{s,t} \mathrel{:=} \int_s^te^{ku}\mathrm{d} z^v(u),
        \quad I\!Z_{s,t} \mathrel{:=} \int_s^t\mathrm{d} z^v(u),
        \quad I\!Z^{*}_{s,t} \mathrel{:=} \int_s^t\mathrm{d} z^{s}(u).
    \end{equation*}

Consequently, we have

.. math::

    \begin{align*}
        &\int_0^t e^{lks} \mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3} I\!E\!Z_t^{m_4} I\!Z_t^{m_5} I\!Z_t^{*m_6}|v_0]\mathrm{d} s\\
        &= \sum_{i_1=0}^{m_4}\sum_{i_2=0}^{m_5}\sum_{i_3=0}^{m_6}\binom{m_4}{i_1}\binom{m_5}{i_2}\binom{m_6}{i_3} \int_0^t e^{lks} \mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3} I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0] \cdot M(i_1,i_2,i_3)\mathrm{d} s.
    \end{align*}

where
:math:`M(i_1,i_2,i_3)\mathrel{:=} \mathbb{E}[I\!E\!Z_{s,t}^{m_4-i_1} I\!Z_{s,t}^{m_5-i_2} I\!Z_{s,t}^{*m_6-i_3}|v_0]`.
We have established that the conditional joint moment
:math:`\mathbb{E}[I\!E\!Z_{s,t}^{m_4}I\!Z_{s,t}^{m_5}I\!Z_{s,t}^{*m_6}|v_0]` can be computed as
the following "polynomial":

.. math::

    \begin{equation}
        \mathbb{E}[I\!E\!Z_{s,t}^{m_4}I\!Z_{s,t}^{m_5}I\!Z_{s,t}^{*m_6}|v_0]
        = \sum_{\boldsymbol{j}}c_{\boldsymbol{j}}e^{j_1kt}t^{j_2}e^{j_3js}s^{j_4}k^{-j_5}\lambda^{j_6}\mu_v^{j_7}\mu_s^{j_8}\sigma_s^{j_9},
    \end{equation}

where :math:`\boldsymbol{j}\mathrel{:=} (j_1,\dots, j_9), j_1,\dots,j_9` are integers,
:math:`c_{\boldsymbol{j}}` represents the corresponding monomial coefficient. For detailed derivations,
please refer to :doc:`../generated/ajdmom.mdl_svcj.ieziziz_mom`.

With above equation, the conditional joint moment can be computed recursively as follows:

.. math::

    \begin{align}
        &\mathbb{E}[I\!E_t^{m_1} I_t^{m_2} I_t^{*m_3} I\!E\!Z_t^{m_4} I\!Z_t^{m_5} I\!Z_t^{*m_6}|v_0]\nonumber\\
        &=\sum_{i_1=0}^{m_4}\sum_{i_2=0}^{m_5}\sum_{i_3=0}^{m_6}\binom{m_4}{i_1}\binom{m_5}{i_2}\binom{m_6}{i_3}
        \sum_{\boldsymbol{j}}c_{\boldsymbol{j}}e^{j_1kt}t^{j_2}k^{-j_5}\lambda^{j_6}\mu_v^{j_7}\mu_s^{j_8}
        \sigma_s^{j_9}F(m_1, m_2, m_3),%\label{eqn:recursive-ieii-ieziziz}
    \end{align}

where

.. math::

    \begin{align*}
        F(m_1,m_2,m_3)
        &\mathrel{:=} \sum_{i=1}^4\left[\frac{m_1(m_1-1)}{2}f_{6i}
          + \frac{m_2(m_2-1)}{2}g_{6i}
          + m_1m_2h_{6i}
          + \frac{m_3(m_3-1)}{2}q_{6i}\right],
    \end{align*}

and the terms :math:`f_{6i}, g_{6i}, h_{6i}, q_{6i}, i=1,2,3,4` are defined as:

.. math::

    \begin{align*}
        f_{61} &\mathrel{:=} \int_0^te^{(j_3+1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-2}I_s^{m_2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times (v_0-\theta),\\
        f_{62} &\mathrel{:=} \int_0^te^{(j_3+2)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-2}I_s^{m_2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \theta,\\
        f_{63} &\mathrel{:=} \int_0^te^{(j_3+1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-1}I_s^{m_2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \sigma_v,\\
        f_{64} &\mathrel{:=} \int_0^te^{(j_3+1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-2}I_s^{m_2}I_s^{*m_3}I\!E\!Z_s^{i_1+1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s,
    \end{align*}

.. math::

    \begin{align*}
        g_{61} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2-2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times (v_0-\theta),\\
        g_{62} &\mathrel{:=} \int_0^te^{(j_3-0)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2-2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \theta,\\
        g_{63} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1+1}I_s^{m_2-2}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \sigma_v,\\
        g_{64} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2-2}I_s^{*m_3}I\!E\!Z_s^{i_1+1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s,
    \end{align*}

.. math::

    \begin{align*}
        h_{61} &\mathrel{:=} \int_0^te^{(j_3+0)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-1}I_s^{m_2-1}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times (v_0-\theta),\\
        h_{62} &\mathrel{:=} \int_0^te^{(j_3+1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-1}I_s^{m_2-1}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \theta,\\
        h_{63} &\mathrel{:=} \int_0^te^{(j_3+0)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-0}I_s^{m_2-1}I_s^{*m_3}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \sigma_v,\\
        h_{64} &\mathrel{:=} \int_0^te^{(j_3+0)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1-1}I_s^{m_2-1}I_s^{*m_3}I\!E\!Z_s^{i_1+1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s,
    \end{align*}

and

.. math::

    \begin{align*}
        q_{61} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3-2}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times (v_0-\theta),\\
        q_{62} &\mathrel{:=} \int_0^te^{(j_3-0)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3-2}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \theta,\\
        q_{63} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1+1}I_s^{m_2}I_s^{*m_3-2}I\!E\!Z_s^{i_1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s \times \sigma_v,\\
        q_{64} &\mathrel{:=} \int_0^te^{(j_3-1)ks}s^{j_4}\mathbb{E}[I\!E_s^{m_1}I_s^{m_2}I_s^{*m_3-2}I\!E\!Z_s^{i_1+1}I\!Z_s^{i_2}I\!Z_s^{*i_3}|v_0]\mathrm{d} s.
    \end{align*}

Before closing this subsection, we highlight that the final expression for the conditional joint moment
takes the form of a polynomial in :math:`v_0-\theta`. Specifically, it can be expressed as:

.. math::

    \begin{equation}%\label{eqn:ieii-ieziziz-polynomial}
        \mathbb{E}[I\!E_t^{m_1}I_t^{m_2}I_t^{*m_3}I\!E\!Z_t^{m_4}I\!Z_t^{m_5}I\!Z_t^{*m_6}|v_0]
        = \sum_{i=0}^{\lfloor (m_1+m_2)/2 \rfloor + \lfloor m_3/2 \rfloor} c_i (v_0-\theta)^i,
    \end{equation}

where :math:`\lfloor x \rfloor` denotes the floor function, i.e., the greatest integer less than
or equal to :math:`x`. Here, with a slight abuse of notation, :math:`c_i` represents the coefficient,
which can be computed via the recursive equation.
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.ito_mom import int_et
from ajdmom.mdl_svcj.ieziziz_mom import moment_ieziziz


def ieziziz_to_ieii_ieziziz(poly):
    # ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    # s = 0 to
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
          'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        key = (k[0], k[2], k[3], 0, 0, 0, k[4], k[5], k[6], k[7])
        poln.add_keyval(key, v)
    return poln


def expand_ieziziz(poly):
    # from ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    # to ['e^{kt}', 't', 'e^{ks}', 's', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(['e^{kt}', 't', 'e^{ks}', 's', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s'])
    for k, v in poly.items():
        for i in range(k[2] + 1):
            bino = math.comb(k[2], i)
            j = k[2] - i
            key = (k[0], i, k[1], j) + k[3:]
            val = bino * ((-1) ** j) * v
            poln.add_keyval(key, val)
    return poln


def int_e_s_poly(c, tp, m1, m2, poly):
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
          'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        n1 = k[0] + m1
        n2 = k[1] + m2
        # \int_0^t e^{n1 ks} s^{n2} ds
        pol1 = int_et(n1, n2)  # with kf = ('e^{kt}', 't', 'k^{-}')
        for K, V in pol1.items():
            # key = [K[0], K[1], K[2] + k[2], k[3], k[4], k[5], k[6], k[7], k[8], k[9]]
            key = list(K)
            key[2] += k[2]
            key.extend(k[3:])
            if tp in [1, 2, 3]: key[tp + 2] += 1
            val = c * v * V
            poln.add_keyval(tuple(key), val)
    return poln


def key_times_poly(k, poly):
    # k: ['e^{kt}', 't', 'e^{ks}', 's', 'k^{-}',
    #     'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    # poly: ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
    #     'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
          'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    for key, val in poly.items():
        knw = (key[0] + k[0],
               key[1] + k[1],
               key[2] + k[4],
               key[3],
               key[4],
               key[5],
               key[6] + k[5],
               key[7] + k[6],
               key[8] + k[7],
               key[9] + k[8])
        poln.add_keyval(knw, val)
    return poln


def recursive_ieii_ieziziz(n1, n2, n3, n4, n5, n6, ieii_ieziziz):
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
          'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poly = Poly()
    poly.set_keyfor(kf)
    # special case
    if n1 == 0 and n2 == 0 and n3 == 0:
        return ieziziz_to_ieii_ieziziz(moment_ieziziz(n4, n5, n6))  # n4 + n5 + n6 > 0
    # typical cases
    for i1 in range(n4 + 1):
        for i2 in range(n5 + 1):
            for i3 in range(n6 + 1):
                bino1 = math.comb(n4, i1)
                bino2 = math.comb(n5, i2)
                bino3 = math.comb(n6, i3)
                bino = bino1 * bino2 * bino3
                pol1 = expand_ieziziz(moment_ieziziz(n4 - i1, n5 - i2, n6 - i3))  # pol1 can not be empty, i.e., {}
                if len(pol1.keys()) == 0:
                    raise Exception("poly is empty")
                # ['e^{kt}', 't', 'e^{ks}', 's', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
                for k, v in pol1.items():
                    if n1 >= 2 and n2 >= 0 and n3 >= 0:
                        c = bino * Fraction(n1 * (n1 - 1), 2) * v
                        pol2 = int_e_s_poly(c, 1, k[2] + 1, k[3], ieii_ieziziz[(n1 - 2, n2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 2, k[2] + 2, k[3], ieii_ieziziz[(n1 - 2, n2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 3, k[2] + 1, k[3], ieii_ieziziz[(n1 - 1, n2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 4, k[2] + 1, k[3], ieii_ieziziz[(n1 - 2, n2, n3, i1 + 1, i2, i3)])
                        poly.merge(key_times_poly(k, pol2))
                    if n1 >= 0 and n2 >= 2 and n3 >= 0:
                        c = bino * Fraction(n2 * (n2 - 1), 2) * v
                        pol2 = int_e_s_poly(c, 1, k[2] - 1, k[3], ieii_ieziziz[(n1, n2 - 2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 2, k[2] - 0, k[3], ieii_ieziziz[(n1, n2 - 2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 3, k[2] - 1, k[3], ieii_ieziziz[(n1 + 1, n2 - 2, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 4, k[2] - 1, k[3], ieii_ieziziz[(n1, n2 - 2, n3, i1 + 1, i2, i3)])
                        poly.merge(key_times_poly(k, pol2))
                    if n1 >= 1 and n2 >= 1 and n3 >= 0:
                        c = bino * n1 * n2 * v
                        pol2 = int_e_s_poly(c, 1, k[2] + 0, k[3], ieii_ieziziz[(n1 - 1, n2 - 1, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 2, k[2] + 1, k[3], ieii_ieziziz[(n1 - 1, n2 - 1, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 3, k[2] + 0, k[3], ieii_ieziziz[(n1 + 0, n2 - 1, n3, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 4, k[2] + 0, k[3], ieii_ieziziz[(n1 - 1, n2 - 1, n3, i1 + 1, i2, i3)])
                        poly.merge(key_times_poly(k, pol2))
                    if n1 >= 0 and n2 >= 0 and n3 >= 2:
                        c = bino * Fraction(n3 * (n3 - 1), 2) * v
                        pol2 = int_e_s_poly(c, 1, k[2] - 1, k[3], ieii_ieziziz[(n1, n2, n3 - 2, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 2, k[2] - 0, k[3], ieii_ieziziz[(n1, n2, n3 - 2, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 3, k[2] - 1, k[3], ieii_ieziziz[(n1 + 1, n2, n3 - 2, i1, i2, i3)])
                        pol2 += int_e_s_poly(c, 4, k[2] - 1, k[3], ieii_ieziziz[(n1, n2, n3 - 2, i1 + 1, i2, i3)])
                        poly.merge(key_times_poly(k, pol2))
    return poly


def moment_ieii_ieziziz(n1, n2, n3, n4, n5, n6):
    """joint conditional moment of :math:`IE_t,I_t,I_t^{*},IEZ_t,IZ_t,IZ_t^{*}|v_0`

    :param integer n1: order of :math:`IE_t`
    :param integer n2: order of :math:`I_t`
    :param integer n3: order of :math:`I_t^{*}`
    :param integer n4: order of :math:`IEZ_t`
    :param integer n5: order of :math:`IZ_t`
    :param integer n6: order of :math:`IZ_t^{*}`
    :return: poly with ``keyfor`` = ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'lmbd','mu_v','mu_s','sigma_s')
    :rtype: Poly
    """
    if n1 < 0 or n2 < 0 or n3 < 0 or n4 < 0 or n5 < 0 or n6 < 0:
        raise ValueError(f'moment_ieii_ieziziz({n1},{n2},{n3},{n4},{n5},{n6}) is called!')
    #
    # ieii_ieziziz: a dict of moments of E[IE_t^n1 I_t^n2 I_t^{*n3} IEZ_t^n4 IZ_t^n5 IZ_t^{*n6}]
    #
    ieii_ieziziz = {}
    #
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
          'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    #
    # special poly constants, analog to 0 and 1
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    P1.set_keyfor(kf)
    #
    # n1 + n2 + n3 + n4 + n5 + n6 = 0: special case
    #
    ieii_ieziziz[(0, 0, 0, 0, 0, 0)] = P1  # equiv to constant 1
    #
    # n1 + n2 + n3 + n4 + n5 + n6 = 1
    #
    ieii_ieziziz[(1, 0, 0, 0, 0, 0)] = P0  # equiv to constant 0
    ieii_ieziziz[(0, 1, 0, 0, 0, 0)] = P0
    ieii_ieziziz[(0, 0, 1, 0, 0, 0)] = P0
    #
    ieii_ieziziz[(0, 0, 0, 1, 0, 0)] = ieziziz_to_ieii_ieziziz(moment_ieziziz(1, 0, 0))
    ieii_ieziziz[(0, 0, 0, 0, 1, 0)] = ieziziz_to_ieii_ieziziz(moment_ieziziz(0, 1, 0))
    ieii_ieziziz[(0, 0, 0, 0, 0, 1)] = ieziziz_to_ieii_ieziziz(moment_ieziziz(0, 0, 1))
    #
    # n1 + n2 + n3 + n4 + n5 + n6 >= 2
    #
    for n in range(2, n1 + n2 + n3 + n4 + n5 + n6):
        for i1 in range(n + 1):
            for i2 in range(n - i1 + 1):
                for i3 in range(n - i1 - i2 + 1):
                    for i4 in range(n - i1 - i2 - i3 + 1):
                        for i5 in range(n - i1 - i2 - i3 - i4 + 1):
                            i6 = n - i1 - i2 - i3 - i4 - i5
                            poly = recursive_ieii_ieziziz(i1, i2, i3, i4, i5, i6, ieii_ieziziz)
                            # poly.remove_zero()
                            ieii_ieziziz[(i1, i2, i3, i4, i5, i6)] = poly
    # the last one
    poly = recursive_ieii_ieziziz(n1, n2, n3, n4, n5, n6, ieii_ieziziz)
    # poly.remove_zero()
    return poly


def poly2num(poly, par):
    # kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma']
    # kf += ['lmbd', 'mu_v', 'mu_s', 'sigma_s']
    v0, k, theta, sigma = par['v0'], par['k'], par['theta'], par['sigma']
    lmbd, mu_v, h = par['lmbd'], par['mu_v'], par['h']
    mu_s, sigma_s = par['mu_s'], par['sigma_s']
    f = 0
    for K, V in poly.items():
        val = math.exp(K[0] * k * h) * (h ** K[1]) / (k ** K[2])
        val *= ((v0 - theta) ** K[3]) * (theta ** K[4]) * (sigma ** K[5])
        val *= (lmbd ** K[6]) * (mu_v ** K[7])
        val *= (mu_s ** K[8]) * (sigma_s ** K[9])
        f += val * V
    return f


def m_ieii_ieziziz(order, par):
    n1, n2, n3, n4, n5, n6 = order
    moment = moment_ieii_ieziziz(n1, n2, n3, n4, n5, n6)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    poly = moment_ieii_ieziziz(1, 1, 1, 2, 1, 1)
    print("poly = ")
    pprint(poly)
    print(f"which is a poly with keyfor = {poly.keyfor}")



"""
Covariance for One-Factor SV with jumps
"""
import math

from ajdmom.poly import Poly
from ajdmom.mdl_1fsv.cov import moment_yy as m_yy
from ajdmom.cpp_mom import mcpp

from ajdmom.mdl_1fsvj.mom import (
    moment_y,
    dpoly,
    poly2num
)


def moment_yy(l1, l2):
    """Moment :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)',
          'lambda', 'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    #
    for i in range(l2 + 1):
        for j in range(l1 + 1):
            coef = math.comb(l2, i) * math.comb(l1, j)
            pol1 = m_yy(j, i)
            # ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)')
            pol2 = mcpp(l2 - i)  # ('lambda*h','mu','sigma')
            pol3 = mcpp(l1 - j)  # ('lambda*h','mu','sigma')
            keyIndexes = [(0, 1, 2, 3, 4, 5, 6, 7, -1, -1, -1), (-1, 0, -1, -1, -1, -1, -1, -1, 0, 1, 2)]
            poln = pol1.mul_poly(pol2 * pol3, keyIndexes, kf)
            poly.merge(coef * poln)
    return poly


def cov_yy(l1, l2):
    """Moment :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    cov = moment_yy(l1, l2) - (moment_y(l1) * moment_y(l2))
    cov.remove_zero()
    return cov


##########
# scalar and (partial) derivative
##########

def cov(l1, l2, par):
    """Covariance in scalar

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param dict par: parameters in dict.
    :return: scalar of the covariance.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    value = poly2num(covariance, par)
    return value


def dcov(l1, l2, par, wrt):
    """Partial derivative of covariance w.r.t. parameter wrt

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    pold = dpoly(covariance, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l1 = 1 if len(args) == 0 else int(args[0])
    l2 = 1 if len(args) <= 1 else int(args[1])
    cov_l1_l2 = cov_yy(l1, l2)
    print(f"cov_yy({l1},{l2}) = \n")
    pprint(cov_l1_l2)
    print(f"\nwhich is a Poly with attribute keyfor =\n{cov_l1_l2.keyfor}")



## Formula Deriviation - Covariance
from ajdmom import mdl_1fsv
from pprint import pprint

cov21 = mdl_1fsv.cov_yy(2, 1)
msg = f"which is a Poly with attribute keyfor =\n{cov21.keyfor}"
print("cov_yy(2,1) = \n")
pprint(cov21)
print(msg)
